@@ -28,6 +28,7 @@ class OAuth2Client(object):
     :param revocation_endpoint_auth_method: client authentication method for
         revocation endpoint.
     :param scope: Scope that you needed to access user resources.
+    :param state: Shared secret to prevent CSRF attack.
     :param redirect_uri: Redirect URI you registered as callback.
     :param code_challenge_method: PKCE method name, only S256 is supported.
     :param token: A dict of token attributes such as ``access_token``,
@@ -48,12 +49,13 @@ class OAuth2Client(object):
     def __init__(self, session, client_id=None, client_secret=None,
                  token_endpoint_auth_method=None,
                  revocation_endpoint_auth_method=None,
-                 scope=None, redirect_uri=None, code_challenge_method=None,
+                 scope=None, state=None, redirect_uri=None, code_challenge_method=None,
                  token=None, token_placement='header', update_token=None, **metadata):
 
         self.session = session
         self.client_id = client_id
         self.client_secret = client_secret
+        self.state = state
 
         if token_endpoint_auth_method is None:
             if client_secret:
@@ -89,6 +91,7 @@ class OAuth2Client(object):
             'refresh_token_request': set(),
             'refresh_token_response': set(),
             'revoke_token_request': set(),
+            'introspect_token_request': set(),
         }
         self._auth_methods = {}
 
@@ -103,7 +106,8 @@ class OAuth2Client(object):
             self._auth_methods[auth.name] = auth
 
     def client_auth(self, auth_method):
-        auth_method = self._auth_methods.get(auth_method, auth_method)
+        if isinstance(auth_method, str) and auth_method in self._auth_methods:
+            auth_method = self._auth_methods[auth_method]
         return self.client_auth_class(
             client_id=self.client_id,
             client_secret=self.client_secret,
@@ -152,7 +156,7 @@ class OAuth2Client(object):
         return uri, state
 
     def fetch_token(self, url=None, body='', method='POST', headers=None,
-                    auth=None, grant_type=None, **kwargs):
+                    auth=None, grant_type=None, state=None, **kwargs):
         """Generic method for fetching an access token from the token endpoint.
 
         :param url: Access Token endpoint URL, if not configured,
@@ -168,10 +172,11 @@ class OAuth2Client(object):
         :param grant_type: Use specified grant_type to fetch token
         :return: A :class:`OAuth2Token` object (a dict too).
         """
+        state = state or self.state
         # implicit  grant_type
         authorization_response = kwargs.pop('authorization_response', None)
         if authorization_response and '#' in authorization_response:
-            return self.token_from_fragment(authorization_response, kwargs.get('state'))
+            return self.token_from_fragment(authorization_response, state)
 
         session_kwargs = self._extract_session_request_params(kwargs)
 
@@ -179,7 +184,7 @@ class OAuth2Client(object):
             grant_type = 'authorization_code'
             params = parse_authorization_code_response(
                 authorization_response,
-                state=kwargs.get('state'),
+                state=state,
             )
             kwargs['code'] = params['code']
 
@@ -202,26 +207,6 @@ class OAuth2Client(object):
             headers=headers, **session_kwargs
         )
 
-    def _fetch_token(self, url, body='', headers=None, auth=None,
-                     method='POST', **kwargs):
-        if method == 'GET':
-            if '?' in url:
-                url = '&'.join([url, body])
-            else:
-                url = '?'.join([url, body])
-            body = ''
-
-        if headers is None:
-            headers = DEFAULT_HEADERS
-
-        resp = self.session.request(
-            method, url, data=body, headers=headers, auth=auth, **kwargs)
-
-        for hook in self.compliance_hook['access_token_response']:
-            resp = hook(resp)
-
-        return self.parse_response_token(resp.json())
-
     def token_from_fragment(self, authorization_response, state=None):
         token = parse_implicit_response(authorization_response, state)
         return self.parse_response_token(token)
@@ -260,23 +245,20 @@ class OAuth2Client(object):
             url, refresh_token=refresh_token, body=body, headers=headers,
             auth=auth, **session_kwargs)
 
-    def _refresh_token(self, url, refresh_token=None, body='', headers=None,
-                       auth=None, **kwargs):
-        resp = self.session.post(
-            url, data=dict(url_decode(body)), headers=headers,
-            auth=auth, **kwargs)
-
-        for hook in self.compliance_hook['refresh_token_response']:
-            resp = hook(resp)
-
-        token = self.parse_response_token(resp.json())
-        if 'refresh_token' not in token:
-            self.token['refresh_token'] = refresh_token
-
-        if callable(self.update_token):
-            self.update_token(self.token, refresh_token=refresh_token)
-
-        return self.token
+    def ensure_active_token(self, token):
+        if not token.is_expired():
+            return True
+        refresh_token = token.get('refresh_token')
+        url = self.metadata.get('token_endpoint')
+        if refresh_token and url:
+            self.refresh_token(url, refresh_token=refresh_token)
+            return True
+        elif self.metadata.get('grant_type') == 'client_credentials':
+            access_token = token['access_token']
+            new_token = self.fetch_token(url, grant_type='client_credentials')
+            if self.update_token:
+                self.update_token(new_token, access_token=access_token)
+            return True
 
     def revoke_token(self, url, token=None, token_type_hint=None,
                      body=None, auth=None, headers=None, **kwargs):
@@ -290,33 +272,35 @@ class OAuth2Client(object):
                      include in the token request. Prefer kwargs over body.
         :param auth: An auth tuple or method as accepted by requests.
         :param headers: Dict to default request headers with.
-        :return: A :class:`OAuth2Token` object (a dict too).
+        :return: Revocation Response
 
         .. _`RFC7009`: https://tools.ietf.org/html/rfc7009
         """
-        if token is None and self.token:
-            token = self.token.get('refresh_token') or self.token.get('access_token')
+        return self._handle_token_hint(
+            'revoke_token_request', url,
+            token=token, token_type_hint=token_type_hint,
+            body=body, auth=auth, headers=headers, **kwargs)
 
-        if body is None:
-            body = ''
+    def introspect_token(self, url, token=None, token_type_hint=None,
+                         body=None, auth=None, headers=None, **kwargs):
+        """Implementation of OAuth 2.0 Token Introspection defined via `RFC7662`_.
 
-        body, headers = prepare_revoke_token_request(
-            token, token_type_hint, body, headers)
+        :param url: Introspection Endpoint, must be HTTPS.
+        :param token: The token to be introspected.
+        :param token_type_hint: The type of the token that to be revoked.
+                                It can be "access_token" or "refresh_token".
+        :param body: Optional application/x-www-form-urlencoded body to add the
+                     include in the token request. Prefer kwargs over body.
+        :param auth: An auth tuple or method as accepted by requests.
+        :param headers: Dict to default request headers with.
+        :return: Introspection Response
 
-        for hook in self.compliance_hook['revoke_token_request']:
-            url, headers, body = hook(url, headers, body)
-
-        if auth is None:
-            auth = self.client_auth(self.revocation_endpoint_auth_method)
-
-        session_kwargs = self._extract_session_request_params(kwargs)
-        return self._revoke_token(
-            url, body, auth=auth, headers=headers, **session_kwargs)
-
-    def _revoke_token(self, url, body=None, auth=None, headers=None, **kwargs):
-        return self.session.post(
-            url, data=dict(url_decode(body)),
-            headers=headers, auth=auth, **kwargs)
+        .. _`RFC7662`: https://tools.ietf.org/html/rfc7662
+        """
+        return self._handle_token_hint(
+            'introspect_token_request', url,
+            token=token, token_type_hint=token_type_hint,
+            body=body, auth=auth, headers=headers, **kwargs)
 
     def register_compliance_hook(self, hook_type, hook):
         """Register a hook for request/response tweaking.
@@ -328,6 +312,7 @@ class OAuth2Client(object):
         * refresh_token_response: invoked before refresh token parsing.
         * protected_request: invoked before making a request.
         * revoke_token_request: invoked before revoking a token.
+        * introspect_token_request: invoked before introspecting a token.
         """
         if hook_type == 'protected_request':
             self.token_auth.hooks.add(hook)
@@ -347,6 +332,66 @@ class OAuth2Client(object):
         description = token.get('error_description', error)
         self.handle_error(error, description)
 
+    @staticmethod
+    def handle_error(error_type, error_description):
+        raise ValueError('{}: {}'.format(error_type, error_description))
+
+    def _fetch_token(self, url, body='', headers=None, auth=None,
+                     method='POST', **kwargs):
+
+        if method.upper() == 'POST':
+            resp = self.session.post(
+                url, data=dict(url_decode(body)),
+                headers=headers, auth=auth, **kwargs)
+        else:
+            if '?' in url:
+                url = '&'.join([url, body])
+            else:
+                url = '?'.join([url, body])
+            resp = self.session.request(method, url, headers=headers, auth=auth, **kwargs)
+
+        for hook in self.compliance_hook['access_token_response']:
+            resp = hook(resp)
+
+        return self.parse_response_token(resp.json())
+
+    def _refresh_token(self, url, refresh_token=None, body='', headers=None,
+                       auth=None, **kwargs):
+        resp = self._http_post(url, body=body, auth=auth, headers=headers, **kwargs)
+
+        for hook in self.compliance_hook['refresh_token_response']:
+            resp = hook(resp)
+
+        token = self.parse_response_token(resp.json())
+        if 'refresh_token' not in token:
+            self.token['refresh_token'] = refresh_token
+
+        if callable(self.update_token):
+            self.update_token(self.token, refresh_token=refresh_token)
+
+        return self.token
+
+    def _handle_token_hint(self, hook, url, token=None, token_type_hint=None,
+                           body=None, auth=None, headers=None, **kwargs):
+        if token is None and self.token:
+            token = self.token.get('refresh_token') or self.token.get('access_token')
+
+        if body is None:
+            body = ''
+
+        body, headers = prepare_revoke_token_request(
+            token, token_type_hint, body, headers)
+
+        for hook in self.compliance_hook[hook]:
+            url, headers, body = hook(url, headers, body)
+
+        if auth is None:
+            auth = self.client_auth(self.revocation_endpoint_auth_method)
+
+        session_kwargs = self._extract_session_request_params(kwargs)
+        return self._http_post(
+            url, body, auth=auth, headers=headers, **session_kwargs)
+
     def _prepare_token_endpoint_body(self, body, grant_type, **kwargs):
         if grant_type is None:
             grant_type = _guess_grant_type(kwargs)
@@ -368,9 +413,10 @@ class OAuth2Client(object):
                 rv[k] = kwargs.pop(k)
         return rv
 
-    @staticmethod
-    def handle_error(error_type, error_description):
-        raise ValueError('{}: {}'.format(error_type, error_description))
+    def _http_post(self, url, body=None, auth=None, headers=None, **kwargs):
+        return self.session.post(
+            url, data=dict(url_decode(body)),
+            headers=headers, auth=auth, **kwargs)
 
 
 def _guess_grant_type(kwargs):
